/*
 *      IP Virtual Server
 *      data structure and functionality definitions
 */

#ifndef _IP_VS_H
#define _IP_VS_H

#include <sys/types.h>

#define IP_VS_VERSION_CODE            0x01000A
#define NVERSION(version)                       \
	(version >> 16) & 0xFF,                 \
	(version >> 8) & 0xFF,                  \
	version & 0xFF

/* Responses from hook functions. */
#define IP_VS_DROP 				0
#define IP_VS_ACCEPT 			1
#define IP_VS_STOLEN 			2
#define IP_VS_QUEUE 			3
#define IP_VS_REPEAT 			4
#define IP_VS_MAX_VERDICT 		NF_REPEAT

/*
 *      Virtual Service Flags
 */
#define IP_VS_SVC_F_PERSISTENT       	0x0001    /* persistent port */
#define IP_VS_SVC_F_HASHED            	0x0002    /* hashed entry */

/*
 *      Destination Server Flags
 */
#define IP_VS_DEST_F_AVAILABLE       	0x0001    /* Available tag */

/*
 *      IPVS sync daemon states
 */
#define IP_VS_STATE_NONE    			0           /* daemon is stopped */
#define IP_VS_STATE_MASTER  		1           /* started as master */
#define IP_VS_STATE_BACKUP  		2           /* started as backup */

/*
 *      IPVS socket options
 */
#define IP_VS_BASE_CTL		(64+1024+64)		/* base */

#define IP_VS_SO_SET_NONE	IP_VS_BASE_CTL	        /* just peek */
#define IP_VS_SO_SET_INSERT	(IP_VS_BASE_CTL+1)
#define IP_VS_SO_SET_ADD	(IP_VS_BASE_CTL+2)
#define IP_VS_SO_SET_EDIT	(IP_VS_BASE_CTL+3)
#define IP_VS_SO_SET_DEL	(IP_VS_BASE_CTL+4)
#define IP_VS_SO_SET_FLUSH	(IP_VS_BASE_CTL+5)
#define IP_VS_SO_SET_LIST	(IP_VS_BASE_CTL+6)
#define IP_VS_SO_SET_ADDDEST	(IP_VS_BASE_CTL+7)
#define IP_VS_SO_SET_DELDEST	(IP_VS_BASE_CTL+8)
#define IP_VS_SO_SET_EDITDEST	(IP_VS_BASE_CTL+9)
#define IP_VS_SO_SET_TIMEOUTS	(IP_VS_BASE_CTL+10)
#define IP_VS_SO_SET_STARTDAEMON (IP_VS_BASE_CTL+11)
#define IP_VS_SO_SET_STOPDAEMON (IP_VS_BASE_CTL+12)
#define IP_VS_SO_SET_RESTORE    (IP_VS_BASE_CTL+13)
#define IP_VS_SO_SET_SAVE       (IP_VS_BASE_CTL+14)
#define IP_VS_SO_SET_ZERO	(IP_VS_BASE_CTL+15)
#define IP_VS_SO_SET_MAX	IP_VS_SO_SET_ZERO

#define IP_VS_SO_GET_VERSION	IP_VS_BASE_CTL
#define IP_VS_SO_GET_INFO	(IP_VS_BASE_CTL+1)
#define IP_VS_SO_GET_SERVICES	(IP_VS_BASE_CTL+2)
#define IP_VS_SO_GET_SERVICE	(IP_VS_BASE_CTL+3)
#define IP_VS_SO_GET_DESTS	(IP_VS_BASE_CTL+4)
#define IP_VS_SO_GET_DEST	(IP_VS_BASE_CTL+5)	/* not used now */
#define IP_VS_SO_GET_TIMEOUTS	(IP_VS_BASE_CTL+6)
#define IP_VS_SO_GET_DAEMON	(IP_VS_BASE_CTL+7)
#define IP_VS_SO_GET_MAX	IP_VS_SO_GET_DAEMON


/*
 *      IPVS Connection Flags
 */
#define IP_VS_CONN_F_FWD_MASK         0x0007    /* mask for the fwd methods */
#define IP_VS_CONN_F_MASQ	      			0x0000    /* masquerading */
#define IP_VS_CONN_F_LOCALNODE	      0x0001    /* local node */
#define IP_VS_CONN_F_TUNNEL	      0x0002    /* tunneling */
#define IP_VS_CONN_F_DROUTE           0x0003    /* direct routing */
#define IP_VS_CONN_F_BYPASS           0x0004    /* cache bypass */
#define IP_VS_CONN_F_HASHED	      0x0040	/* hashed entry */
#define IP_VS_CONN_F_NOOUTPUT         0x0080    /* no output packets */
#define IP_VS_CONN_F_INACTIVE         0x0100    /* not established */
#define IP_VS_CONN_F_OUT_SEQ          0x0200    /* must do output seq adjust */
#define IP_VS_CONN_F_IN_SEQ           0x0400    /* must do input seq adjust */
#define IP_VS_CONN_F_SEQ_MASK         0x0600    /* in/out sequence mask */
#define IP_VS_CONN_F_NO_CPORT         0x0800    /* no client port set yet */

/* Move it to better place one day, for now keep it unique */
#define NFC_IPVS_PROPERTY	0x10000

#define IP_VS_SCHEDNAME_MAXLEN         16
#define IP_VS_IFNAME_MAXLEN            16

struct ip_vs_rule_user {
	/* global options */
	int             tcp_timeout;    /* timeout values */
	int             tcp_fin_timeout;
	int             udp_timeout;
	int             state;          /* sync daemon state */
	char            mcast_ifn[IP_VS_IFNAME_MAXLEN];
					/* multicast interface name */

	/* virtual service options */
	u_int16_t		protocol;
	u_int32_t		vaddr;          /* virtual address */
	u_int16_t		vport;
	u_int32_t       vfwmark;        /* firwall mark of virtual service*/
	char            	sched_name[IP_VS_SCHEDNAME_MAXLEN];
	unsigned		vs_flags;       /* virtual service flags */
	unsigned        timeout;        /* persistent timeout in ticks */
	u_int32_t		netmask;        /* persistent netmask */

	/* destination specific options */
	u_int32_t		daddr;          /* destination address */
	u_int16_t		dport;
	unsigned        conn_flags;     /* destination flags */
	int             	weight;         /* destination weight */
};


/*
 *	IPVS statistics object (for user space)
 */
struct ip_vs_stats_user
{
	u_int32_t                   conns;          /* connections scheduled */
	u_int32_t                   inpkts;         /* incoming packets */
	u_int32_t                   outpkts;        /* outgoing packets */
	u_int64_t                   inbytes;        /* incoming bytes */
	u_int64_t                   outbytes;       /* outgoing bytes */

	u_int32_t			cps;		/* current connection rate */
	u_int32_t			inpps;		/* current in packet rate */
	u_int32_t			outpps;		/* current out packet rate */
	u_int32_t			inbps;		/* current in byte rate */
	u_int32_t			outbps;		/* current out byte rate */
};


/* The argument to IP_VS_SO_GET_INFO */
struct ip_vs_getinfo {
	/* version number */
	unsigned int	version;

	/* size of connection hash table */
	unsigned int	size;

	/* number of virtual services */
	unsigned int	num_services;
};

/* The argument to IP_VS_SO_GET_SERVICE */
struct ip_vs_service_user {
	/* which service: user fills this in */
	u_int16_t		protocol;
	u_int32_t		addr;           /* virtual address */
	u_int16_t		port;
	u_int32_t       fwmark;         /* firwall mark of virtual service */

	/* service options */
	char            sched_name[IP_VS_SCHEDNAME_MAXLEN];
	unsigned		flags;          /* virtual service flags */
	unsigned        timeout;        /* persistent timeout in ticks */
	u_int32_t		netmask;        /* persistent netmask */

	/* number of real servers */
	unsigned int    num_dests;

	/* statistics */
	struct ip_vs_stats_user stats;
};

struct ip_vs_dest_user {
	u_int32_t		addr;           /* destination address */
	u_int16_t		port;
	unsigned		flags;		/* destination flags */
	int			weight;         /* destination weight */
	u_int32_t		activeconns;	/* active connections */
	u_int32_t		inactconns;	/* inactive connections */

	/* statistics */
	struct ip_vs_stats_user stats;
};

/* The argument to IP_VS_SO_GET_DESTS */
struct ip_vs_get_dests {
	/* which service: user fills this in */
	u_int16_t		protocol;
	u_int32_t		addr;           /* virtual address */
	u_int16_t		port;
	u_int32_t       fwmark;         /* firwall mark of virtual service */

	/* number of real servers */
	unsigned int    num_dests;

	/* the real servers */
	struct ip_vs_dest_user entrytable[0];
};

/* The argument to IP_VS_SO_GET_SERVICES */
struct ip_vs_get_services {
	/* number of virtual services */
	unsigned int num_services;

	/* service table */
	struct ip_vs_service_user entrytable[0];
};

/* The argument to IP_VS_SO_GET_TIMEOUTS */
struct ip_vs_timeout_user {
	int             tcp_timeout;
	int             tcp_fin_timeout;
	int             udp_timeout;
};

/* The argument to IP_VS_SO_GET_DAEMON */
struct ip_vs_daemon_user {
	int		state;				/* sync daemon state */
	char		mcast_ifn[IP_VS_IFNAME_MAXLEN];	/* multicast interface name */
};

#ifdef _KERNEL

#define NF_SOCKOPT

#include <sys/cdefs.h>
#include <sys/param.h>
#include <sys/module.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/sysproto.h>
#include <sys/fcntl.h>
#include <sys/file.h>
#include <sys/limits.h>
#include <sys/malloc.h>
#include <sys/queue.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/syscallsubr.h>
#include <sys/uio.h>
#include <sys/syslog.h>
#include <sys/smp.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/sx.h>

#include <net/if.h>
#include <net/route.h>
#include <net/pfil.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>

#include <machine/in_cksum.h>

#include "linux_jhash.h"

static MALLOC_DEFINE(M_IPVSBUF, "ipvsbuf", "Buffers for the IPVS");

int net_ratelimit(void);

#ifdef CONFIG_IP_VS_DEBUG
extern int ip_vs_get_debug_level(void);
#define IP_VS_DBG(level, msg...)			\
    do {						\
	    if (level <= ip_vs_get_debug_level())	\
		    log(LOG_DEBUG, "IPVS: " msg);	\
    } while (0)
#define IP_VS_DBG_RL(msg...)				\
    do {						\
	    if (net_ratelimit())			\
		    log(LOG_DEBUG, "IPVS: " msg);	\
    } while (0)
#else	/* NO DEBUGGING at ALL */
#define IP_VS_DBG(level, msg...)  do {} while (0)
#define IP_VS_DBG_RL(msg...)  do {} while (0)
#endif

#define IP_VS_BUG() BUG()
#define IP_VS_ERR(msg...) 	log(LOG_ERR, "IPVS: " msg)
#define IP_VS_INFO(msg...)	log(LOG_INFO, "IPVS: " msg)
#define IP_VS_WARNING(msg...) \
	log(LOG_WARNING, "IPVS: " msg)
#define IP_VS_ERR_RL(msg...)				\
    do {						\
	    if (net_ratelimit())			\
		    log(LOG_ERR, "IPVS: " msg);	\
    } while (0)

#ifdef CONFIG_IP_VS_DEBUG
#define EnterFunction(level)						\
    do {								\
	    if (level <= ip_vs_get_debug_level())			\
		    log(LOG_DEBUG, "Enter: %s, %s line %i\n",	\
			   __FUNCTION__, __FILE__, __LINE__);		\
    } while (0)
#define LeaveFunction(level)                                            \
    do {                                                                \
	    if (level <= ip_vs_get_debug_level())                       \
			log(LOG_DEBUG, "Leave: %s, %s line %i\n",    \
			       __FUNCTION__, __FILE__, __LINE__);       \
    } while (0)
#else
#define EnterFunction(level)   do {} while (0)
#define LeaveFunction(level)   do {} while (0)
#endif


/*
 *      The port number of FTP service (in network order).
 */
#define FTPPORT  			htons(21)
#define FTPDATA  			htons(20)

#define PROT_SOCK			1024

/*
 *      IPVS sysctl variables under the /proc/sys/net/ipv4/vs/
 */
#define NET_IPV4_VS              21

enum {
	NET_IPV4_VS_DEBUG_LEVEL=1,
	NET_IPV4_VS_AMEMTHRESH=2,
	NET_IPV4_VS_AMDROPRATE=3,
	NET_IPV4_VS_DROP_ENTRY=4,
	NET_IPV4_VS_DROP_PACKET=5,
	NET_IPV4_VS_SECURE_TCP=6,
	NET_IPV4_VS_TO_ES=7,
	NET_IPV4_VS_TO_SS=8,
	NET_IPV4_VS_TO_SR=9,
	NET_IPV4_VS_TO_FW=10,
	NET_IPV4_VS_TO_TW=11,
	NET_IPV4_VS_TO_CL=12,
	NET_IPV4_VS_TO_CW=13,
	NET_IPV4_VS_TO_LA=14,
	NET_IPV4_VS_TO_LI=15,
	NET_IPV4_VS_TO_SA=16,
	NET_IPV4_VS_TO_UDP=17,
	NET_IPV4_VS_TO_ICMP=18,
	NET_IPV4_VS_LBLC_EXPIRE=19,
	NET_IPV4_VS_LBLCR_EXPIRE=20,
	NET_IPV4_VS_CACHE_BYPASS=22,
	NET_IPV4_VS_EXPIRE_NODEST_CONN=23,
	NET_IPV4_VS_SYNC_THRESHOLD=24,
	NET_IPV4_VS_NAT_ICMP_SEND=25,
	NET_IPV4_VS_LAST
};


/*
 *      IPVS State Values
 */
enum {
	IP_VS_S_NONE = 0,
	IP_VS_S_ESTABLISHED,
	IP_VS_S_SYN_SENT,
	IP_VS_S_SYN_RECV,
	IP_VS_S_FIN_WAIT,
	IP_VS_S_TIME_WAIT,
	IP_VS_S_CLOSE,
	IP_VS_S_CLOSE_WAIT,
	IP_VS_S_LAST_ACK,
	IP_VS_S_LISTEN,
	IP_VS_S_SYNACK,
	IP_VS_S_UDP,
	IP_VS_S_ICMP,
	IP_VS_S_LAST
};


struct ip_vs_timeout_table {
	volatile u_int32_t refcnt;
	int scale;
	int timeout[IP_VS_S_LAST+1];
};


/*
 *	Transport protocol header
 */
union ip_vs_tphdr {
	unsigned char *raw;
	struct udphdr *uh;
	struct tcphdr *th;
	struct icmp *icmph;
	u_int16_t *portp;
};


/*
 *	Delta sequence info structure
 *	Each ip_vs_conn has 2 (output AND input seq. changes).
 *      Only used in the VS/NAT.
 */
struct ip_vs_seq {
	u_int32_t           init_seq;       /* Add delta from this seq */
	u_int32_t           delta;          /* Delta in sequence numbers */
	u_int32_t           previous_delta; /* Delta in sequence numbers
					   before last resized pkt */
};


/*
 *	IPVS statistics object
 */
struct ip_vs_stats
{
	u_int32_t                   conns;          /* connections scheduled */
	u_int32_t                   inpkts;         /* incoming packets */
	u_int32_t                   outpkts;        /* outgoing packets */
	u_int64_t                   inbytes;        /* incoming bytes */
	u_int64_t                   outbytes;       /* outgoing bytes */

	u_int32_t			cps;		/* current connection rate */
	u_int32_t			inpps;		/* current in packet rate */
	u_int32_t			outpps;		/* current out packet rate */
	u_int32_t			inbps;		/* current in byte rate */
	u_int32_t			outbps;		/* current out byte rate */

	struct mtx              lock;           /* spin lock */
};

typedef int ipvs_hookfn(void *arg,
			       struct mbuf **m,
			       struct ifnet *in,
			       int flags,
			       struct inpcb *inp);

struct ipvs_hook_ops
{
	ipvs_hookfn *hook;
	int pf;
	int hookpos;
};

/*
 *	IP_VS structure allocated for each dynamically scheduled connection
 */
struct ip_vs_conn {
	TAILQ_ENTRY(ip_vs_conn) 	c_list;	/* hashed list entry */

	/* Protocol, addresses and port numbers */
	u_int32_t                   caddr;          /* client address */
	u_int32_t                   vaddr;          /* virtual address */
	u_int32_t                   daddr;          /* destination address */
	u_int16_t                   cport;
	u_int16_t                   vport;
	u_int16_t                   dport;
	u_int16_t                   protocol;       /* Which protocol (TCP/UDP) */

	/* counter and timer */
	volatile u_int32_t		refcnt;		/* reference count */
	struct callout			timer;		/* Expiration timer */
	volatile unsigned long	timeout;	/* timeout */
	struct ip_vs_timeout_table *timeout_table;

	/* Flags and state transition */
	struct mtx              lock;           /* lock for state transition */
	volatile u_int16_t          flags;          /* status flags */
	volatile u_int16_t          state;          /* state info */

	/* Control members */
	struct ip_vs_conn       *control;       /* Master control connection */
	volatile u_int32_t                n_control;      /* Number of controlled ones */
	struct ip_vs_dest       *dest;          /* real server */
	volatile u_int32_t                in_pkts;        /* incoming packet counter */

	/* packet transmitter for different forwarding methods */
	int (*packet_xmit)(struct mbuf *skb, struct ip_vs_conn *cp);

};


/*
 *	The information about the virtual service offered to the net
 *	and the forwarding entries
 */
struct ip_vs_service {
	LIST_ENTRY(ip_vs_service) 	s_list;	/* for normal service table */
	LIST_ENTRY(ip_vs_service) 	f_list;	/* for fwmark-based service table */
	volatile u_int32_t		refcnt;   /* reference counter */
	volatile u_int32_t		usecnt;   /* use counter */

	u_int16_t			protocol; /* which protocol (TCP/UDP) */
	u_int32_t			addr;	  /* IP address for virtual service */
	u_int16_t			port;	  /* port number for the service */
	u_int32_t                   fwmark;   /* firewall mark of the service */
	unsigned		flags;	  /* service status flags */
	unsigned		timeout;  /* persistent timeout in ticks */
	u_int32_t			netmask;  /* grouping granularity */

	LIST_HEAD(ip_vs_dest_list, ip_vs_dest)	destinations;	 /* real server d-linked list */
	u_int32_t			num_dests;     /* number of servers */
	struct ip_vs_stats      stats;         /* statistics for the service */

	/* for scheduling */
	struct ip_vs_scheduler	*scheduler;    /* bound scheduler object */
	struct sx		sched_lock;    /* lock sched_data */
	void			*sched_data;   /* scheduler application data */
};


/*
 *	The real server destination forwarding entry
 *	with ip address, port number, and so on.
 */
struct ip_vs_dest {
	LIST_ENTRY(ip_vs_dest) n_list; /* for the dests in the service */
	LIST_ENTRY(ip_vs_dest) d_list; /* for table with all the dests */

	u_int32_t			addr;	  /* IP address of real server */
	u_int16_t			port;	  /* port number of the service */
	unsigned		flags;	  /* dest status flags */
	volatile u_int32_t		weight;	  /* server weight */
	volatile u_int32_t		conn_flags;	/* flags to copy to conn */
	volatile u_int32_t		activeconns;	/* active connections */
	volatile u_int32_t		inactconns;     /* inactive connections */
	volatile u_int32_t		refcnt;	        /* reference counter */
	struct ip_vs_stats      stats;          /* statistics */

	/* for destination cache */
	struct mtx		dst_lock;	/* lock dst_cache */
	struct route	dst_cache;	/* destination cache entry */

	/* for virtual service */
	struct ip_vs_service    *svc;     /* service that it belongs to */
	u_int16_t			protocol; /* which protocol (TCP/UDP) */
	u_int32_t			vaddr;	  /* IP address for virtual service */
	u_int16_t			vport;	  /* port number for the service */
	u_int32_t                   vfwmark;  /* firewall mark of the service */
};


/*
 *	The scheduler object
 */
struct ip_vs_scheduler {
	SLIST_ENTRY(ip_vs_scheduler) n_list; /* d-linked list head */
	char			*name;    /* scheduler name */
	volatile u_int32_t                refcnt;   /* reference counter */
	struct module		*module;  /* THIS_MODULE/NULL */

	/* scheduler initializing service */
	int (*init_service)(struct ip_vs_service *svc);
	/* scheduling service finish */
	int (*done_service)(struct ip_vs_service *svc);
	/* scheduler updating service */
	int (*update_service)(struct ip_vs_service *svc);

	/* selecting a server from the given service */
	struct ip_vs_dest* (*schedule)(struct ip_vs_service *svc,
				       struct ip *iph);
};


/*
 *      IPVS core functions
 *      (from ip_vs_core.c)
 */
static inline const char *ip_vs_proto_name(unsigned proto)
{
	static char buf[20];

	switch (proto) {
	case IPPROTO_IP:
		return "IP";
	case IPPROTO_UDP:
		return "UDP";
	case IPPROTO_TCP:
		return "TCP";
	case IPPROTO_ICMP:
		return "ICMP";
	default:
		sprintf(buf, "IP_%d", proto);
		return buf;
	}
}

/*
 *     ip_vs_conn handling functions
 *     (from ip_vs_conn.c)
 */

/*
 *     IPVS connection entry hash table
 */
#ifndef CONFIG_IP_VS_TAB_BITS
#define CONFIG_IP_VS_TAB_BITS   12
#endif
/* make sure that IP_VS_CONN_TAB_BITS is located in [8, 20] */
#if CONFIG_IP_VS_TAB_BITS < 8
#define IP_VS_CONN_TAB_BITS	8
#endif
#if CONFIG_IP_VS_TAB_BITS > 20
#define IP_VS_CONN_TAB_BITS	20
#endif
#if 8 <= CONFIG_IP_VS_TAB_BITS && CONFIG_IP_VS_TAB_BITS <= 20
#define IP_VS_CONN_TAB_BITS	CONFIG_IP_VS_TAB_BITS
#endif
#define IP_VS_CONN_TAB_SIZE     (1 << IP_VS_CONN_TAB_BITS)
#define IP_VS_CONN_TAB_MASK     (IP_VS_CONN_TAB_SIZE - 1)

#define VS_STATE_INPUT	        0
#define VS_STATE_OUTPUT	        4
#define VS_STATE_INPUT_ONLY	8

extern struct ip_vs_timeout_table vs_timeout_table;
extern struct ip_vs_timeout_table vs_timeout_table_dos;

extern struct ip_vs_conn *ip_vs_conn_in_get
(int protocol, u_int32_t s_addr, u_int16_t s_port, u_int32_t d_addr, u_int16_t d_port);

#define	ENTRY_ON_LIST(elm, field) 	((elm)->field.le_prev != NULL)

#define	INIT_LIST_ENTRY(elm, field) do {						\
	(elm)->field.le_prev = NULL;					\
	(elm)->field.le_next = NULL;			\
} while (0)

#define atomic_read(p)		(*(p))
#define atomic_set(p, v)	(*(p) = (v))

static inline u_int atomic_dec_and_test(volatile u_int32_t *p)
{
	atomic_subtract_int(p, 1);
	//atomic_dec(p);
	return atomic_read(p)==0;
}

#define NIPQUAD(addr) \
	((unsigned char *)&addr)[0], \
	((unsigned char *)&addr)[1], \
	((unsigned char *)&addr)[2], \
	((unsigned char *)&addr)[3]

enum
{
	RTN_UNSPEC,
	RTN_UNICAST,			/* Gateway or direct route	*/
	RTN_BROADCAST,		/* Accept locally as broadcast,
				   				send as broadcast */
	RTN_MULTICAST,			/* Multicast route		*/
};

#define LOOPBACK(x)	(((x) & htonl(0xff000000)) == htonl(0x7f000000))
#define MULTICAST(x)	(((x) & htonl(0xf0000000)) == htonl(0xe0000000))
#define BADCLASS(x)	(((x) & htonl(0xf0000000)) == htonl(0xf0000000))
#define ZERONET(x)	(((x) & htonl(0xff000000)) == htonl(0x00000000))
#define LOCAL_MCAST(x)	(((x) & htonl(0xFFFFFF00)) == htonl(0xE0000000))

static inline unsigned inet_addr_type(u_int32_t addr)
{
	unsigned ret = RTN_UNICAST;

	if (ZERONET(addr) || BADCLASS(addr))
		return RTN_BROADCAST;
	if (MULTICAST(addr))
		return RTN_MULTICAST;
	return ret;
}

#define HZ	hz

/* put back the conn without restarting its timer */
static inline void __ip_vs_conn_put(struct ip_vs_conn *cp)
{
	atomic_subtract_int(&cp->refcnt, 1);
}
extern void ip_vs_conn_put(struct ip_vs_conn *cp);

extern struct ip_vs_conn *
ip_vs_conn_new(int proto, u_int32_t caddr, u_int16_t cport, u_int32_t vaddr, u_int16_t vport,
	       u_int32_t daddr, u_int16_t dport, unsigned flags,
	       struct ip_vs_dest *dest);
extern void ip_vs_conn_expire_now(struct ip_vs_conn *cp);

extern const char * ip_vs_state_name(int state);
extern int ip_vs_set_state(struct ip_vs_conn *cp, int state_off,
			   struct ip *iph, void *tp);
extern int ip_vs_conn_listen(struct ip_vs_conn *cp);
extern int ip_vs_check_template(struct ip_vs_conn *ct);
extern void ip_vs_secure_tcp_set(int on);
extern int ip_vs_conn_init(void);
extern void ip_vs_conn_cleanup(void);

static inline void ip_vs_control_del(struct ip_vs_conn *cp)
{
	struct ip_vs_conn *ctl_cp = cp->control;
	if (!ctl_cp) {
		IP_VS_ERR("request control DEL for uncontrolled: "
			  "%u.%u.%u.%u:%d to %u.%u.%u.%u:%d\n",
			  NIPQUAD(cp->caddr),ntohs(cp->cport),
			  NIPQUAD(cp->vaddr),ntohs(cp->vport));
		return;
	}

	IP_VS_DBG(7, "DELeting control for: "
		  "cp.dst=%u.%u.%u.%u:%d ctl_cp.dst=%u.%u.%u.%u:%d\n",
		  NIPQUAD(cp->caddr),ntohs(cp->cport),
		  NIPQUAD(ctl_cp->caddr),ntohs(ctl_cp->cport));

	cp->control = NULL;
	if (atomic_read(&ctl_cp->n_control) == 0) {
		IP_VS_ERR("BUG control DEL with n=0 : "
			  "%u.%u.%u.%u:%d to %u.%u.%u.%u:%d\n",
			  NIPQUAD(cp->caddr),ntohs(cp->cport),
			  NIPQUAD(cp->vaddr),ntohs(cp->vport));
		return;
	}
	atomic_subtract_int(&ctl_cp->n_control, 1);
}

static inline void
ip_vs_control_add(struct ip_vs_conn *cp, struct ip_vs_conn *ctl_cp)
{
	if (cp->control) {
		IP_VS_ERR("request control ADD for already controlled: "
			  "%u.%u.%u.%u:%d to %u.%u.%u.%u:%d\n",
			  NIPQUAD(cp->caddr),ntohs(cp->cport),
			  NIPQUAD(cp->vaddr),ntohs(cp->vport));
		ip_vs_control_del(cp);
	}

	IP_VS_DBG(7, "ADDing control for: "
		  "cp.dst=%u.%u.%u.%u:%d ctl_cp.dst=%u.%u.%u.%u:%d\n",
		  NIPQUAD(cp->caddr),ntohs(cp->cport),
		  NIPQUAD(ctl_cp->caddr),ntohs(ctl_cp->cport));

	cp->control = ctl_cp;
	atomic_add_int(&ctl_cp->n_control, 1);
}


/*
 *      Registering/unregistering scheduler functions
 *      (from ip_vs_sched.c)
 */
extern void ip_vs_sched_init(void);
extern void ip_vs_sched_cleanup(void);
extern int register_ip_vs_scheduler(struct ip_vs_scheduler *scheduler);
extern int unregister_ip_vs_scheduler(struct ip_vs_scheduler *scheduler);
extern int ip_vs_bind_scheduler(struct ip_vs_service *svc,
				struct ip_vs_scheduler *scheduler);
extern int ip_vs_unbind_scheduler(struct ip_vs_service *svc);
extern struct ip_vs_scheduler *ip_vs_scheduler_get(const char *sched_name);
extern void ip_vs_scheduler_put(struct ip_vs_scheduler *scheduler);

/*
  *	   Round Robin schedule functions
  *	    (from ip_vs_rr.c)
  */
extern int ip_vs_rr_init(void);
extern void ip_vs_rr_cleanup(void);

/*
 *      IPVS control data and functions
 *      (from ip_vs_ctl.c)
 */
extern int sysctl_ip_vs_cache_bypass;
extern int sysctl_ip_vs_expire_nodest_conn;
extern int sysctl_ip_vs_sync_threshold;
extern struct ip_vs_stats ip_vs_stats;

extern struct ip_vs_service *ip_vs_service_get(u_int32_t fwmark,
					       u_int16_t protocol,
					       u_int32_t vaddr, u_int16_t vport);
static inline void ip_vs_service_put(struct ip_vs_service *svc)
{
	atomic_subtract_int(&svc->usecnt, 1);
}

extern struct ip_vs_dest *
ip_vs_lookup_real_service(u_int16_t protocol, u_int32_t daddr, u_int16_t dport);
extern int ip_vs_control_init(void);
extern void ip_vs_control_cleanup(void);

/*
 *      ip_vs_fwd_tag returns the forwarding tag of the connection
 */
#define IP_VS_FWD_METHOD(cp)  (cp->flags & IP_VS_CONN_F_FWD_MASK)

static __inline__ char ip_vs_fwd_tag(struct ip_vs_conn *cp)
{
	char fwd;

	switch (IP_VS_FWD_METHOD(cp)) {
	case IP_VS_CONN_F_MASQ:
		fwd = 'M'; break;
	case IP_VS_CONN_F_LOCALNODE:
		fwd = 'L'; break;
	case IP_VS_CONN_F_TUNNEL:
		fwd = 'T'; break;
	case IP_VS_CONN_F_DROUTE:
		fwd = 'R'; break;
	case IP_VS_CONN_F_BYPASS:
		fwd = 'B'; break;
	default:
		fwd = '?'; break;
	}
	return fwd;
}


/*
 *	transport layer header checking
 */
static inline int ip_vs_header_check(struct mbuf **skb, int proto, int ihl)
{
	int len;
	struct mbuf *m=*skb;
	
	switch (proto) {
	case IPPROTO_TCP:
		len = ihl + sizeof(struct tcphdr);
		/* we don't care about TCP options */
		break;
	case IPPROTO_UDP:
		len = ihl + sizeof(struct udphdr);
		break;
	default:
		len = 0;
	}

	/* guarantee protocol header available in skb data area */
	if (m->m_len < len) {
		if ((m = m_pullup(m,len)) == NULL)
			return -1;
		*skb = m;
	}
	return 0;
}

static inline void ip_rt_put(struct route *ro)
{
	if (ro->ro_rt) {
		RTFREE(ro->ro_rt);
		ro->ro_rt = NULL;
	}
}

static inline void
__ip_vs_dst_release(struct ip_vs_dest *dest)
{
	struct route *rt = &dest->dst_cache;

	ip_rt_put(rt);
}

static inline int
__ip_vs_rt_check(struct route *ro)
{
	struct sockaddr_in *dst;

	if (!ro->ro_rt)
		return -1;
	dst = (struct sockaddr_in *)&ro->ro_dst;
	/*
	 * If there is a cached route,
	 * check that it is to the same destination
	 * and is still up.  If not, free it and try again.
	 * The address family should also be checked in case of sharing the
	 * cache with IPv6.
	 */
	if ((ro->ro_rt->rt_flags & RTF_UP) == 0) {
		ip_rt_put(ro);
		return -1;
	}
	return 0;
}

static inline int 
__ip_vs_get_out_rt(struct ip_vs_conn *cp)
{
	struct ip_vs_dest *dest = cp->dest;
	struct route *ro;
	int ret = -1;
	
	if (!dest)
		goto out;
	mtx_lock(&dest->dst_lock);
	ro = &dest->dst_cache;
	if (!__ip_vs_rt_check(ro)) { 
		ret = 0;
		goto out_unlock;
	}
	rtalloc(ro);
	if (ro->ro_rt) 
		ret = 0;
	out_unlock:
	mtx_unlock(&dest->dst_lock);
	out:
	return ret;
}

static inline int
skb_cow(struct mbuf **skb, unsigned int headroom)
{
	struct mbuf *m=*skb;
	int delta = (headroom > 16 ? headroom : 16);

	M_PREPEND(m, (delta+15)&~15, M_DONTWAIT);
	if (!m) {
		log( LOG_ERR,"%s: M_PREPEND failed\n",  __FUNCTION__);
		return ENOBUFS;
	}

	*skb = m;
	
	return 0;
}

static inline int
ip_vs_skb_cow(struct mbuf **skb, unsigned int headroom,
	      struct ip **iph_p, unsigned char **t_p)
{
	struct mbuf *m=*skb;
	
	int delta = (headroom > 16 ? headroom : 16);

	M_PREPEND(m, (delta+15)&~15, M_DONTWAIT);
	if (!m) {
		log( LOG_ERR,"%s: M_PREPEND failed\n", __FUNCTION__);
		return ENOBUFS;
	}
	*skb = m;

	/* skb data changed, update pointers */
	*iph_p = mtod(m,struct ip *);
	*t_p = (char*) (*iph_p) + (*iph_p)->ip_hl * 4;
	return 0;
}

#endif  	/* _KERNEL */

#endif	/* _IP_VS_H */
