/*
 * IPVS         An implementation of the IP virtual server support for the
 *              LINUX operating system.  IPVS is now implemented as a module
 *              over the Netfilter framework. IPVS can be used to build a
 *              high-performance and highly available server based on a
 *              cluster of servers.
 *
 * Version:     $Id: ip_vs_core.c,v 1.31.2.5 2003/07/29 14:37:12 wensong Exp $
 *
 * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
 *              Peter Kese <peter.kese@ijs.si>
 *              Julian Anastasov <ja@ssi.bg>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * The IPVS code for kernel 2.2 was done by Wensong Zhang and Peter Kese,
 * with changes/fixes from Julian Anastasov, Lars Marowsky-Bree, Horms
 * and others.
 *
 * Changes:
 *
 */

/*
  *	Port IPVS to FreeBSD
  *	Authors:		Li Wang <dragonfly@linux-vs.org>
  *
  */

#include "ip_vs.h"  

static inline void
ip_vs_in_stats(struct ip_vs_conn *cp, struct mbuf *skb)
{
	struct ip_vs_dest *dest = cp->dest;
	if (dest && (dest->flags & IP_VS_DEST_F_AVAILABLE)) {
		mtx_lock(&dest->stats.lock);
		dest->stats.inpkts++;
		dest->stats.inbytes += skb->m_pkthdr.len;
		mtx_unlock(&dest->stats.lock);

		mtx_lock(&dest->svc->stats.lock);
		dest->svc->stats.inpkts++;
		dest->svc->stats.inbytes += skb->m_pkthdr.len;
		mtx_unlock(&dest->svc->stats.lock);

		mtx_lock(&ip_vs_stats.lock);
		ip_vs_stats.inpkts++;
		ip_vs_stats.inbytes += skb->m_pkthdr.len;
		mtx_unlock(&ip_vs_stats.lock);
	}
}

static inline void
ip_vs_conn_stats(struct ip_vs_conn *cp, struct ip_vs_service *svc)
{
	mtx_lock(&cp->dest->stats.lock);
	cp->dest->stats.conns++;
	mtx_unlock(&cp->dest->stats.lock);

	mtx_lock(&svc->stats.lock);
	svc->stats.conns++;
	mtx_unlock(&svc->stats.lock);

	mtx_lock(&ip_vs_stats.lock);
	ip_vs_stats.conns++;
	mtx_unlock(&ip_vs_stats.lock);
}

/*
 *  IPVS persistent scheduling function
 *  It creates a connection entry according to its template if exists,
 *  or selects a server and creates a connection entry plus a template.
 *  Locking: we are svc user (svc->refcnt), so we hold all dests too
 */
static struct ip_vs_conn *
ip_vs_sched_persist(struct ip_vs_service *svc, struct ip *iph)
{
	struct ip_vs_conn *cp = NULL;
	struct ip_vs_dest *dest;
	const u_int16_t *portp;
	struct ip_vs_conn *ct;
	u_int16_t  dport;	 /* destination port to forward */
	u_int32_t  snet;	 /* source network of the client, after masking */

	portp = (u_int16_t *)&(((char *)iph)[iph->ip_hl*4]);

	/* Mask saddr with the netmask to adjust template granularity */
	snet = iph->ip_src.s_addr & svc->netmask;

	IP_VS_DBG(6, "P-schedule: src %u.%u.%u.%u:%u dest %u.%u.%u.%u:%u "
		  "mnet %u.%u.%u.%u\n",
		  NIPQUAD(iph->saddr), ntohs(portp[0]),
		  NIPQUAD(iph->daddr), ntohs(portp[1]),
		  NIPQUAD(snet));

	/*
	 * As far as we know, FTP is a very complicated network protocol, and
	 * it uses control connection and data connections. For active FTP,
	 * FTP server initialize data connection to the client, its source port
	 * is often 20. For passive FTP, FTP server tells the clients the port
	 * that it passively listens to,  and the client issues the data
	 * connection. In the tunneling or direct routing mode, the load
	 * balancer is on the client-to-server half of connection, the port
	 * number is unknown to the load balancer. So, a conn template like
	 * <caddr, 0, vaddr, 0, daddr, 0> is created for persistent FTP
	 * service, and a template like <caddr, 0, vaddr, vport, daddr, dport>
	 * is created for other persistent services.
	 */
	if (portp[1] == svc->port) {
		/* Check if a template already exists */
		if (svc->port != FTPPORT)
			ct = ip_vs_conn_in_get(iph->ip_p, snet, 0,
					       iph->ip_dst.s_addr, portp[1]);
		else
			ct = ip_vs_conn_in_get(iph->ip_p, snet, 0,
					       iph->ip_dst.s_addr, 0);

		if (!ct || !ip_vs_check_template(ct)) {
			/*
			 * No template found or the dest of the connection
			 * template is not available.
			 */
			dest = svc->scheduler->schedule(svc, iph);
			if (dest == NULL) {
				IP_VS_DBG(1, "P-schedule: no dest found.\n");
				return NULL;
			}

			/*
			 * Create a template like <protocol,caddr,0,
			 * vaddr,vport,daddr,dport> for non-ftp service,
			 * and <protocol,caddr,0,vaddr,0,daddr,0>
			 * for ftp service.
			 */
			if (svc->port != FTPPORT)
				ct = ip_vs_conn_new(iph->ip_p,
						    snet, 0,
						    iph->ip_dst.s_addr, portp[1],
						    dest->addr, dest->port,
						    0,
						    dest);
			else
				ct = ip_vs_conn_new(iph->ip_p,
						    snet, 0,
						    iph->ip_dst.s_addr, 0,
						    dest->addr, 0,
						    0,
						    dest);
			if (ct == NULL)
				return NULL;

			ct->timeout = svc->timeout;
		} else {
			/* set destination with the found template */
			dest = ct->dest;
		}
		dport = dest->port;
	} else {
		/*
		 * Note: persistent fwmark-based services and persistent
		 * port zero service are handled here.
		 * fwmark template: <IPPROTO_IP,caddr,0,fwmark,0,daddr,0>
		 * port zero template: <protocol,caddr,0,vaddr,0,daddr,0>
		 */
		if (svc->fwmark)
			ct = ip_vs_conn_in_get(IPPROTO_IP, snet, 0,
					       htonl(svc->fwmark), 0);
		else
			ct = ip_vs_conn_in_get(iph->ip_p, snet, 0,
					       iph->ip_dst.s_addr, 0);

		if (!ct || !ip_vs_check_template(ct)) {
			/*
			 * If it is not persistent port zero, return NULL,
			 * otherwise create a connection template.
			 */
			if (svc->port)
				return NULL;

			dest = svc->scheduler->schedule(svc, iph);
			if (dest == NULL) {
				IP_VS_DBG(1, "P-schedule: no dest found.\n");
				return NULL;
			}

			/*
			 * Create a template according to the service
			 */
			if (svc->fwmark)
				ct = ip_vs_conn_new(IPPROTO_IP,
						    snet, 0,
						    htonl(svc->fwmark), 0,
						    dest->addr, 0,
						    0,
						    dest);
			else
				ct = ip_vs_conn_new(iph->ip_p,
						    snet, 0,
						    iph->ip_dst.s_addr, 0,
						    dest->addr, 0,
						    0,
						    dest);
			if (ct == NULL)
				return NULL;

			ct->timeout = svc->timeout;
		} else {
			/* set destination with the found template */
			dest = ct->dest;
		}
		dport = portp[1];
	}

	/*
	 *    Create a new connection according to the template
	 */
	cp = ip_vs_conn_new(iph->ip_p,
			    iph->ip_src.s_addr, portp[0],
			    iph->ip_dst.s_addr, portp[1],
			    dest->addr, dport,
			    0,
			    dest);
	if (cp == NULL) {
		ip_vs_conn_put(ct);
		return NULL;
	}

	/*
	 *    Increase the inactive connection counter
	 *    because it is in Syn-Received
	 *    state (inactive) when the connection is created.
	 */
	atomic_add_int(&dest->inactconns, 1);

	/*
	 *    Add its control
	 */
	ip_vs_control_add(cp, ct);

	ip_vs_conn_put(ct);
	return cp;
}


/*
 *  IPVS main scheduling function
 *  It selects a server according to the virtual service, and
 *  creates a connection entry.
 */
static struct ip_vs_conn *
ip_vs_schedule(struct ip_vs_service *svc, struct ip *iph)
{
	struct ip_vs_conn *cp = NULL;
	struct ip_vs_dest *dest;
	const u_int16_t *portp;

	/*
	 *    Persistent service
	 */
	if (svc->flags & IP_VS_SVC_F_PERSISTENT)
		return ip_vs_sched_persist(svc, iph);

	/*
	 *    Non-persistent service
	 */
	portp = (u_int16_t *)&(((char *)iph)[iph->ip_hl*4]);
	if (!svc->fwmark && portp[1] != svc->port) {
		if (!svc->port)
			IP_VS_ERR("Schedule: port zero only supported "
				  "in persistent services, "
				  "check your ipvs configuration\n");
		return NULL;
	}

	dest = svc->scheduler->schedule(svc, iph);
	if (dest == NULL) {
		IP_VS_DBG(1, "Schedule: no dest found.\n");
		return NULL;
	}

	/*
	 *    Create a connection entry.
	 */
	cp = ip_vs_conn_new(iph->ip_p,
			    iph->ip_src.s_addr, portp[0],
			    iph->ip_dst.s_addr, portp[1],
			    dest->addr, dest->port?dest->port:portp[1],
			    0,
			    dest);
	if (cp == NULL)
		return NULL;

	/*
	 *    Increase the inactive connection counter because it is in
	 *    Syn-Received state (inactive) when the connection is created.
	 */
	atomic_add_int(&dest->inactconns, 1);

	IP_VS_DBG(6, "Schedule fwd:%c s:%s c:%u.%u.%u.%u:%u v:%u.%u.%u.%u:%u "
		  "d:%u.%u.%u.%u:%u flg:%X cnt:%d\n",
		  ip_vs_fwd_tag(cp), ip_vs_state_name(cp->state),
		  NIPQUAD(cp->caddr), ntohs(cp->cport),
		  NIPQUAD(cp->vaddr), ntohs(cp->vport),
		  NIPQUAD(cp->daddr), ntohs(cp->dport),
		  cp->flags, cp->refcnt);

	return cp;
}


/*
 *  Pass or drop the packet.
 *  Called by ip_vs_in, when the virtual service is available but
 *  no destination is available for a new connection.
 */
static int ip_vs_leave(struct ip_vs_service *svc, struct mbuf *skb)
{
	struct ip *iph = mtod(skb,struct ip *);
	u_int16_t *portp = (u_int16_t *)&(((char *)iph)[iph->ip_hl*4]);

	/* if it is fwmark-based service, the cache_bypass sysctl is up
	   and the destination is RTN_UNICAST (and not local), then create
	   a cache_bypass connection entry */
	if (sysctl_ip_vs_cache_bypass && svc->fwmark) {
		int ret;
		struct ip_vs_conn *cp;

		ip_vs_service_put(svc);

		/* create a new connection entry */
		IP_VS_DBG(6, "ip_vs_leave: create a cache_bypass entry\n");
		cp = ip_vs_conn_new(iph->ip_p,
				    iph->ip_src.s_addr, portp[0],
				    iph->ip_dst.s_addr, portp[1],
				    0, 0,
				    IP_VS_CONN_F_BYPASS,
				    NULL);
		if (cp == NULL) {
			return IP_VS_DROP;
		}

		/* statistics */
		ip_vs_in_stats(cp, skb);
		
		/* set state */
		ip_vs_set_state(cp, VS_STATE_INPUT, iph, portp);

		/* transmit the first SYN packet */
		ret = cp->packet_xmit(skb, cp);

		atomic_add_int(&cp->in_pkts, 1);
		ip_vs_conn_put(cp);
		return ret;
	}

	/*
	 * When the virtual ftp service is presented, packets destined
	 * for other services on the VIP may get here (except services
	 * listed in the ipvs table), pass the packets, because it is
	 * not ipvs job to decide to drop the packets.
	 */
	if ((svc->port == FTPPORT) && (portp[1] != FTPPORT)) {
		ip_vs_service_put(svc);
		return IP_VS_ACCEPT;
	}

	ip_vs_service_put(svc);

	/*
	 * Notify the client that the destination is unreachable, and
	 * release the socket buffer.
	 * Since it is in IP layer, the TCP socket is not actually
	 * created, the TCP RST packet cannot be sent, instead that
	 * ICMP_PORT_UNREACH is sent here no matter it is TCP/UDP. --WZ
	 */
	icmp_error(skb, ICMP_UNREACH, ICMP_UNREACH_PORT, 0, NULL);
	return IP_VS_STOLEN;
}


/*
 *	Handle ICMP messages in the outside-to-inside direction (incoming)
 *	and sometimes in outgoing direction from ip_vs_forward_icmp.
 *	Find any that might be relevant, check against existing connections,
 *	forward to the right destination host if relevant.
 *	Currently handles error types - unreachable, quench, ttl exceeded.
 */
static int ip_vs_in_icmp(struct mbuf **skb_p)
{
	struct mbuf	*skb   = *skb_p;
	u_int ihl;
	struct ip    *iph;
	struct icmp  *icmph;
	struct ip    *ciph;	/* The ip header contained within the ICMP */
	u_int16_t	        *pptr;	/* port numbers from TCP/UDP contained header */
	unsigned short   len;
	unsigned short	clen, csize;
	struct ip_vs_conn *cp;

	
	iph = mtod(skb,struct ip *);
	ihl = iph->ip_hl << 2;
	skb->m_len -= ihl;
	if (skb->m_len < sizeof(struct icmp)) {
		if ((skb = m_pullup(skb,sizeof(struct icmp))) == NULL) {
			return IP_VS_DROP;
		}
		*skb_p=skb;
	}
	skb->m_data += ihl;
	icmph = mtod(skb,struct icmp *);
	len = iph->ip_len - ihl;
	if (len < sizeof(struct icmp))
		return IP_VS_DROP;
	skb->m_len += ihl;
	skb->m_data -= ihl;
	IP_VS_DBG(12, "icmp in (%d,%d) %u.%u.%u.%u -> %u.%u.%u.%u\n",
		  icmph->icmp_type,ntohs(icmph->icmp_id),
		  NIPQUAD(iph->ip_src.s_addr), NIPQUAD(iph->ip_dst.s_addr));
	if ((icmph->icmp_type != ICMP_UNREACH) &&
	    (icmph->icmp_type != ICMP_SOURCEQUENCH) &&
	    (icmph->icmp_type != ICMP_TIMXCEED))
		return IP_VS_ACCEPT;
	/*
	 * If we get here we have an ICMP error of one of the above 3 types
	 * Now find the contained IP header
	 */
	clen = len - sizeof(struct icmp);
	if (clen < sizeof(struct ip))
		return IP_VS_DROP;
	ciph = (struct ip *) (icmph + 1);
	csize = ciph->ip_hl << 2;
	if (clen < csize)
		return IP_VS_DROP;

	/* We are only interested ICMPs generated from TCP or UDP packets */
	if (ciph->ip_p != IPPROTO_UDP && ciph->ip_p != IPPROTO_TCP)
		return IP_VS_ACCEPT;

	/* Skip non-first embedded TCP/UDP fragments */
	/* Note:ip_off has been changed to host byte order in FreeBSD */
	if (ciph->ip_off & IP_OFFMASK)
		return IP_VS_ACCEPT;

	/* We need at least TCP/UDP ports here */
	if (clen < csize + sizeof(struct udphdr))
		return IP_VS_DROP;

	/* Ensure the checksum is correct */
	skb->m_len -= ihl;
	skb->m_data += ihl;
	if (in_cksum(skb,len)) {
		/* Failed checksum! */
		IP_VS_ERR_RL("incoming ICMP: failed checksum from "
			     "%u.%u.%u.%u!\n", NIPQUAD(iph->ip_src.s_addr));
		return IP_VS_DROP;
	}
	skb->m_len += ihl;
	skb->m_data -= ihl;
	pptr = (u_int16_t *)&(((char *)ciph)[csize]);

	IP_VS_DBG(11, "Handling incoming ICMP for "
		  "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
		  NIPQUAD(ciph->ip_src.s_addr), ntohs(pptr[0]),
		  NIPQUAD(ciph->ip_dst.s_addr), ntohs(pptr[1]));

	/* This is pretty much what ip_vs_conn_in_get() does,
	   except parameters are in the reverse order */
	cp = ip_vs_conn_in_get(ciph->ip_p,
			       ciph->ip_dst.s_addr, pptr[1],
			       ciph->ip_src.s_addr, pptr[0]);
	if (cp == NULL)
		return IP_VS_ACCEPT;

	/* statistics */
	ip_vs_in_stats(cp, skb);
	
	/* The ICMP packet for VS/TUN, VS/DR and LOCALNODE will be
	   forwarded directly here, because there is no need to
	   translate address/port back */
	if (IP_VS_FWD_METHOD(cp) != IP_VS_CONN_F_MASQ) {
		int ret;
		if (cp->packet_xmit) {
			ret = cp->packet_xmit(skb, cp);
		}
		else
			ret = IP_VS_ACCEPT;
		atomic_add_int(&cp->in_pkts, 1);
		ip_vs_conn_put(cp);
		return ret;
	}

	ip_vs_conn_put(cp);
	return IP_VS_DROP;
}

/*
 *	Check if it's for virtual services, look it up,
 *	and send it on its way...
 */
static int __ip_vs_in(void *arg,
			     struct mbuf **skb_p,
			     struct ifnet *in,
			     int flags,
			     struct inpcb *inp)
{
	struct mbuf	*skb = *skb_p;
	struct ip *iph = mtod(skb,struct ip *);
	union ip_vs_tphdr h;
	struct ip_vs_conn *cp;
	struct ip_vs_service *svc;
	int ihl;
	int ret;

	/*
	 *	Big tappo: only PACKET_HOST (nor loopback neither mcasts)
	 *	... don't know why 1st test DOES NOT include 2nd (?)
	 */
	if ((skb->m_pkthdr.rcvif->if_flags & IFF_LOOPBACK) ) {
		return IP_VS_ACCEPT;
	}

	if (skb->m_flags & (M_MCAST|M_BCAST)) {
		return IP_VS_ACCEPT;
	}
	
	if (iph->ip_p == IPPROTO_ICMP)
		return ip_vs_in_icmp(skb_p);
	
	/* let it go if other IP protocols */
	if (iph->ip_p != IPPROTO_TCP && iph->ip_p != IPPROTO_UDP)
		return IP_VS_ACCEPT;
	
	/* make sure that protocol header available in skb data area,
	   note that skb data area may be reallocated. */
	ihl = iph->ip_hl << 2;
	if (ip_vs_header_check(&skb, iph->ip_p, ihl) == -1)
		return IP_VS_DROP;
	
	iph = mtod(skb,struct ip *);
	h.raw = (char*) iph + ihl;

	/*
	 * Check if the packet belongs to an existing connection entry
	 */
	cp = ip_vs_conn_in_get(iph->ip_p, iph->ip_src.s_addr, h.portp[0],
			       iph->ip_dst.s_addr, h.portp[1]);

	if (!cp &&
	    (h.th->th_flags & TH_SYN || (iph->ip_p!=IPPROTO_TCP)) &&
	    (svc = ip_vs_service_get(0, iph->ip_p,
				     iph->ip_dst.s_addr, h.portp[1]))) {
		/*
		 * Let the virtual server select a real server for the
		 * incoming connection, and create a connection entry.
		 */
		cp = ip_vs_schedule(svc, iph);
		if (!cp)
			return ip_vs_leave(svc, skb);
		ip_vs_conn_stats(cp, svc);
		ip_vs_service_put(svc);
	}

	if (!cp) {
		/* sorry, all this trouble for a no-hit :) */
		IP_VS_DBG(12, "packet for %s %u.%u.%u.%u:%d continue "
			  "traversal as normal.\n",
			  ip_vs_proto_name(iph->ip_p),
			  NIPQUAD(iph->ip_dst.s_addr),
			  ntohs(h.portp[1]));
		return IP_VS_ACCEPT;
	}
	IP_VS_DBG(11, "Incoming %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d\n",
		  ip_vs_proto_name(iph->ip_p),
		  NIPQUAD(iph->ip_src.s_addr), ntohs(h.portp[0]),
		  NIPQUAD(iph->ip_dst.s_addr), ntohs(h.portp[1]));
	/* Check the server status */
	if (cp->dest && !(cp->dest->flags & IP_VS_DEST_F_AVAILABLE)) {
		/* the destination server is not availabe */
		if (sysctl_ip_vs_expire_nodest_conn) {
			/* try to expire the connection immediately */
			ip_vs_conn_expire_now(cp);
		} else {
			/* don't restart its timer, and silently
			   drop the packet. */
			__ip_vs_conn_put(cp);
		}
		return IP_VS_DROP;
	}

	/* statistics */
	ip_vs_in_stats(cp, skb);

	ip_vs_set_state(cp, VS_STATE_INPUT, iph, h.portp);
	if (cp->packet_xmit) {
		ret = cp->packet_xmit(skb, cp);
	}
	else {
		IP_VS_DBG_RL("warning: packet_xmit is null");
		ret = IP_VS_ACCEPT;
	}

	/* increase its packet counter and check if it is needed
	   to be synchronized */
	atomic_add_int(&cp->in_pkts, 1);
	ip_vs_conn_put(cp);
	return ret;
}

static int ip_vs_in(void *arg,
			     struct mbuf **skb_p,
			     struct ifnet *in,
			     int flags,
			     struct inpcb *inp)
{
	switch (__ip_vs_in(arg,skb_p,in,flags,inp)) {
		case IP_VS_DROP:
			m_freem(*skb_p);
		case IP_VS_STOLEN:
			*skb_p=NULL;
			break;
		default:
			break;
	}
	return 0;
}

int ipvs_register_hook(struct ipvs_hook_ops *);

int ipvs_register_hook(struct ipvs_hook_ops *reg)
{
	struct pfil_head *pfh_inet;
	int flags = PFIL_WAITOK;

	if (reg->pf != AF_INET && reg->pf != PF_INET) 
		return EPROTONOSUPPORT;
	flags |= reg->hookpos;
	pfh_inet = pfil_head_get(PFIL_TYPE_AF, AF_INET);
	if (pfh_inet == NULL)
		return ENOENT;
	return pfil_add_hook(reg->hook, NULL, flags, pfh_inet);
}

void ipvs_unregister_hook(struct ipvs_hook_ops *);

void ipvs_unregister_hook(struct ipvs_hook_ops *reg)
{
	struct pfil_head *pfh_inet;
	int flags = PFIL_WAITOK;
	
	if (reg->pf!=AF_INET && reg->pf != PF_INET) {
		IP_VS_ERR("Unregister unkown ipvs_hook\n");
		return;
	}
	flags |= reg->hookpos;
	pfh_inet = pfil_head_get(PFIL_TYPE_AF, AF_INET);
	if (pfh_inet == NULL) {
		IP_VS_ERR("Packet hook list not found\n");
		return;
	}
	if (pfil_remove_hook(reg->hook, NULL, flags, pfh_inet))
		IP_VS_ERR("Remove hook error\n");
}

/* After packet filtering, forward packet through VS/DR, VS/TUN,
   or VS/NAT(change destination), so that filtering rules can be
   applied to IPVS. */
static struct ipvs_hook_ops ip_vs_in_ops = {
	ip_vs_in, PF_INET, PFIL_IN
};

/*
 *	Initialize IP Virtual Server
 */
static int ip_vs_init(void)
{
	int ret;

	ip_vs_sched_init();
	
	ret = ip_vs_control_init();
	if (ret < 0) {
		IP_VS_ERR("can't setup control.\n");
		goto cleanup_nothing;
	}

	ret = ip_vs_conn_init();
	if (ret < 0) {
		IP_VS_ERR("can't setup connection table.\n");
		goto cleanup_control;
	}

	ret = ipvs_register_hook(&ip_vs_in_ops);
	if (ret < 0) {
		IP_VS_ERR("can't register in hook.\n");
		goto cleanup_conn;
	}

	ret=ip_vs_rr_init();
	if (ret < 0) {
		IP_VS_ERR("can't register round_robin scheduler.\n");
		goto cleanup_inops;
	}
	
	IP_VS_INFO("ipvs loaded.\n");
	return ret;

  cleanup_inops:
	ipvs_unregister_hook(&ip_vs_in_ops);
  cleanup_conn:
	ip_vs_conn_cleanup();
  cleanup_control:
	ip_vs_control_cleanup();
  cleanup_nothing:
	return ret;
}

static void ip_vs_cleanup(void)
{
	ip_vs_rr_cleanup();
	ipvs_unregister_hook(&ip_vs_in_ops);
	ip_vs_conn_cleanup();
	ip_vs_control_cleanup();
	ip_vs_sched_cleanup();
	IP_VS_INFO("ipvs unloaded.\n");
}

/*
 * Load handler that deals with the loading and unloading of a KLD.
 */
static int ipvs_loader(struct module *m, int what, void *arg)
{
 	int err = 0;
	
    	switch (what) {
    	case MOD_LOAD:
		err=ip_vs_init();
        	break;
    	case MOD_UNLOAD:
	case MOD_SHUTDOWN:
		ip_vs_cleanup();
        	break;
    	default:
		err = EOPNOTSUPP;
      		break;
    	}

    	return (err);
}

static moduledata_t ipvs_mod = {
    "ipvs",
    ipvs_loader,
    NULL
};

DECLARE_MODULE(ipvs, ipvs_mod,SI_SUB_KLD, SI_ORDER_ANY);
